//
// Copyright (c) 2011, Franz Haring
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//

using System;
using System.Collections;

namespace de.itools.opensource.utils {
  /// <summary>
  /// Klasse zur Verarbeitung einer Kommandozeile.
  /// </summary>
  /// Die Klasse bietet Zugriffsmethoden auf einzelne Elemente wie
  /// Optionen oder Aufrufparameter.
  public class CommandLine {
    /// <summary>
    /// Konstruktor.
    /// </summary>
    /// <param name="sProgramPath">Pfad und Dateiname der ausfhrbaren Datei, die 
    /// gestartet wurde.</param>
    /// <param name="args">Die Aufrufargumente. Beispiel: "/v -f=out.txt in.txt".</param>
    /// <param name="bCaseSensitive">Wenn true, dann werden die Optionen nach 
    /// Gro- und Kleinschreibung unterschieden.</param>
    public CommandLine(string sProgramPath,string[] args,bool bCaseSensitive) {
      this.sProgramPath=sProgramPath;
      this.args=args;
      this.bCaseSensitive=bCaseSensitive;
      this.parseArgs(args);
    }

    /// <summary>
    /// Den Wert einer Option zurckliefern.
    /// Wenn die Aufrufargumente z.B. "/v -f=out.txt in.txt" waren,
    /// dann liefert getOption("f") den Wert "out.txt" zurck, getOption("v")
    /// liefert "" und getOption("x") liefert null zurck.
    /// </summary>
    /// <param name="sOptionKey">Der Optionsname; bei Optionen mit Werten der Teil
    /// links vom '='-Zeichen.</param>
    /// <returns>Der Optionswert; bei Optionen mit Werten der Teil
    /// rechts vom '='-Zeichen; bei Optionen ohne Werte ein Leerstring;
    /// bei nicht vorhandenen Optionen null.</returns>
    public string GetOption(string sOptionKey) {
      return this.GetOption(sOptionKey,null);
    }

    /// <summary>
    /// Den Wert einer Option oder einen Standardwert zurckliefern.
    /// </summary>
    /// <param name="sOptionKey">Der Optionsname; bei Optionen mit Werten der Teil
    /// links vom '='-Zeichen.</param>
    /// <param name="sDefaultValue">Der Standardwert; wird bei nicht vorhandenen
    /// Optionen zurckgeliefert.</param>
    /// <returns>Der Optionswert; bei Optionen mit Werten der Teil
    /// rechts vom '='-Zeichen; bei Optionen ohne Werte ein Leerstring;
    /// bei nicht vorhandenen Optionen sDefaultValue.</returns>
    public string GetOption(string sOptionKey,string sDefaultValue) {
      string sOptionValue=(string)this.htOptions[this.bCaseSensitive?sOptionKey:sOptionKey.ToUpper()];
      if (sOptionValue==null) {
        return sDefaultValue;
      } else {
        return sOptionValue;
      }
    }

    /// <summary>
    /// Liefert alle Optionsnamen zurck.
    /// </summary>
    public string[] OptionKeys {
      get {
        string[] sOptionKeys=new string[this.htOptions.Count];
        this.htOptions.Keys.CopyTo(sOptionKeys,0);
        return sOptionKeys;
      }
    }

    /// <summary>
    /// Liefert alle Aufrufargumente auer den Optionen zurck.
    /// </summary>
    public string[] Items {
      get {
        string[] sItems=new string[this.alItems.Count];
        this.alItems.CopyTo(sItems,0);
        return sItems;
      }
    }

    protected string sProgramPath;
    protected string[] args;
    protected bool bCaseSensitive;

    protected ArrayList alItems=new ArrayList();
    protected Hashtable htOptions=new Hashtable();

    protected void parseArgs(string[] args) {
      bool is_windows = true;
      switch (System.Environment.OSVersion.Platform) {
        case PlatformID.Unix:
          is_windows = false;
          break;
        case PlatformID.MacOSX:
          is_windows = false;
          break;
      }

      for (int iArg=0;iArg<args.Length;iArg++) {
        string sArg=args[iArg];
        if ((is_windows && sArg.StartsWith("/")) || sArg.StartsWith("-")) {
          int nIdxEquals=sArg.IndexOf('=');
          string sOptionKey;
          string sOptionValue;
          if (nIdxEquals<1) {
            sOptionKey=sArg.Substring(1);
            sOptionValue="";
          } else {
            sOptionKey=sArg.Substring(1,nIdxEquals-1);
            sOptionValue=sArg.Substring(nIdxEquals+1);
          }
          this.htOptions[this.bCaseSensitive?sOptionKey:sOptionKey.ToUpper()]=sOptionValue;
        } else {
          this.alItems.Add(sArg);
        }
      }
    }
  }
}
